import numpy as np
from pyopu.tensors.base_tensor import Rank1Tensor, Rank2Tensor, Rank3Tensor
from pyopu.telemetry.base_telemetry import Observable

class OPUResult:
    def __init__(self, final_probabilities):
        self.final_probabilities = final_probabilities

    def get_binary_state(self, threshold=0.5):
        return [1 if p > threshold else 0 for p in self.final_probabilities]

class TensorEngine(Observable):
    def __init__(self, organoid, interface):
        super().__init__()
        self.organoid = organoid
        self.interface = interface
        self.T1 = None; self.T2 = None; self.T3 = None

    def set_tensors(self, t1: Rank1Tensor, t2: Rank2Tensor = None, t3: Rank3Tensor = None):
        nv = t1.num_vars
        if t2 is not None and t2.num_vars != nv: raise ValueError("Rank-2 var mismatch.")
        if t3 is not None and t3.num_vars != nv: raise ValueError("Rank-3 var mismatch.")
        self.T1 = t1.data
        self.T2 = t2.data if t2 is not None else None
        self.T3 = t3.data if t3 is not None else None

    def _calculate_force(self):
        Force = np.copy(self.T1)
        if self.T2 is not None:
            Force += np.dot(self.T2, self.organoid.x_state) + np.dot(self.T2.T, self.organoid.x_state)
        if self.T3 is not None:
            F3 = np.einsum('ijk,j,k->i', self.T3, self.organoid.x_state, self.organoid.x_state)
            F3 += np.einsum('jik,j,k->i', self.T3, self.organoid.x_state, self.organoid.x_state)
            F3 += np.einsum('jki,j,k->i', self.T3, self.organoid.x_state, self.organoid.x_state)
            Force += F3
        return Force

    def run(self, duration_ms=1000.0, dt_kernel_ms=10.0):
        if self.T1 is None: raise ValueError("Tensors not set.")
        steps = int(duration_ms / dt_kernel_ms)
        print(f"Starting OPU Simulation ({self.organoid.num_neobits} Neobits)...")

        for step in range(steps):
            x_inst = self.interface.read_from_organoid(self.organoid._control_spike_mon, self.organoid.neurons_per_neobit, self.organoid.num_neobits)
            self.organoid.x_state = 0.8 * self.organoid.x_state + 0.2 * x_inst
            force = self._calculate_force()
            self.interface.write_to_organoid(self.organoid.neuron_group, force, self.organoid.neurons_per_neobit)
            
            # Notify observers (like EnergyTelemetry)
            self.notify_telemetries(x_state=self.organoid.x_state, force=force)
            
            self.organoid.run_step(dt_kernel_ms)
            
        print("Execution Complete.")
        return OPUResult(self.organoid.x_state)



import numpy as np
from .base_telemetry import BaseTelemetry

class EnergyTelemetry(BaseTelemetry):
    def __init__(self):
        self.history_energy = []

    def attach(self, engine):
        pass

    def update(self, x_state, force):
        energy = -np.sum(x_state * force)
        self.history_energy.append(energy)

    def get_data(self):
        return {"energy_over_time": self.history_energy}



from abc import ABC, abstractmethod

class BaseTelemetry(ABC):
    @abstractmethod
    def attach(self, subject):
        pass

    def update(self, *args, **kwargs):
        pass
        
    @abstractmethod
    def get_data(self):
        pass

class Observable:
    def __init__(self):
        self._telemetries = []

    def attach_telemetry(self, telemetry: BaseTelemetry):
        self._telemetries.append(telemetry)
        telemetry.attach(self)

    def notify_telemetries(self, *args, **kwargs):
        for telemetry in self._telemetries:
            telemetry.update(*args, **kwargs)



import numpy as np

class Tensor:
    def __init__(self, data, expected_rank):
        self._data = np.array(data, dtype=np.float64)
        self.rank = expected_rank
        if self._data.ndim != self.rank:
            raise ValueError(f"Expected Rank-{self.rank}, got dim={self._data.ndim}.")
        self.shape = self._data.shape
        if self.rank > 1 and not all(dim == self.shape[0] for dim in self.shape):
            raise ValueError(f"Rank-{self.rank} Tensor must be hypercubic. Shape: {self.shape}.")
        self.num_vars = self.shape[0] if self.shape else 0

    @property
    def data(self):
        return self._data

class Rank1Tensor(Tensor):
    def __init__(self, data):
        super().__init__(data, expected_rank=1)

class Rank2Tensor(Tensor):
    def __init__(self, data):
        super().__init__(data, expected_rank=2)

class Rank3Tensor(Tensor):
    def __init__(self, data):
        super().__init__(data, expected_rank=3)



import numpy as np
import brian2 as b2
from .base_interface import BaseInterface

class OptoInterface(BaseInterface):
    def __init__(self, temperature=1.0, max_rate_hz=200.0):
        self.T_sys = temperature
        self.max_rate_hz = max_rate_hz
        self.last_spike_count = None

    def get_biological_equations(self):
        return '''
        I_stim = I_blue + I_yellow : amp
        I_blue = 2.0 * nA * O_blue * (0*mV - v) / (60*mV) : amp
        I_yellow = 2.0 * nA * O_yellow * (-90*mV - v) / (60*mV) : amp
        dO_blue/dt = (phi_blue * (1 - O_blue) - 0.1/ms * O_blue) : 1
        dO_yellow/dt = (phi_yellow * (1 - O_yellow) - 0.05/ms * O_yellow) : 1
        phi_blue : Hz
        phi_yellow : Hz
        '''

    def read_from_organoid(self, spike_monitor, neurons_per_neobit, num_neobits):
        if self.last_spike_count is None: self.last_spike_count = np.zeros(num_neobits * neurons_per_neobit)
        curr = np.array(spike_monitor.count)
        delta = curr - self.last_spike_count
        self.last_spike_count = curr
        x_inst = np.zeros(num_neobits)
        for i in range(num_neobits):
            if np.sum(delta[i*neurons_per_neobit : (i+1)*neurons_per_neobit]) > 1: x_inst[i] = 1.0
        return x_inst

    def write_to_organoid(self, neuron_group, force_array, neurons_per_neobit):
        num_neobits = len(force_array)
        prob = 1.0 / (1.0 + np.exp(force_array / self.T_sys))
        blue = np.zeros(num_neobits); yellow = np.zeros(num_neobits)
        for i in range(num_neobits):
            if prob[i] > 0.5: blue[i] = (prob[i] - 0.5) * 2.0
            else: yellow[i] = (0.5 - prob[i]) * 2.0
        neuron_group.phi_blue = np.repeat(blue, neurons_per_neobit) * self.max_rate_hz * b2.Hz
        neuron_group.phi_yellow = np.repeat(yellow, neurons_per_neobit) * self.max_rate_hz * b2.Hz
